from typing import Dict, List, Callable

from yxdb import _extractors
from yxdb._metainfo_field import MetaInfoField
from yxdb.yxdb_field import YxdbField, DataType


class YxdbRecord:
    def __init__(self, fields: List[MetaInfoField]):
        self._name_to_index: Dict[str, int] = {}
        self.fields: List[YxdbField] = []
        self._extractors: List[Callable] = []
        self.has_var = False
        self.fixed_size = 0

        self._initialize(fields)

    def extract_from_index(self, index: int, buffer: memoryview):
        if index < 0 or index >= len(self._extractors):
            raise Exception(f"index {index} is not a valid field index")
        return self._extractors[index](buffer)

    def extract_from_name(self, name: str, buffer: memoryview):
        if name not in self._name_to_index:
            raise Exception(f"'{name}' is not a valid field name")
        index = self._name_to_index[name]
        return self._extractors[index](buffer)

    def _initialize(self, fields: List[MetaInfoField]):
        start_at = 0
        for field in fields:
            if field.data_type == 'Int16':
                self._add_extractor(field.name, DataType.LONG, _extractors.new_int16_extractor(start_at))
                start_at += 3
                continue
            if field.data_type == 'Int32':
                self._add_extractor(field.name, DataType.LONG, _extractors.new_int32_extractor(start_at))
                start_at += 5
                continue
            if field.data_type == 'Int64':
                self._add_extractor(field.name, DataType.LONG, _extractors.new_int64_extractor(start_at))
                start_at += 9
                continue
            if field.data_type == 'Float':
                self._add_extractor(field.name, DataType.DOUBLE, _extractors.new_float_extractor(start_at))
                start_at += 5
                continue
            if field.data_type == 'Double':
                self._add_extractor(field.name, DataType.DOUBLE, _extractors.new_double_extractor(start_at))
                start_at += 9
                continue
            if field.data_type == 'FixedDecimal':
                size = field.size
                self._add_extractor(field.name, DataType.DOUBLE, _extractors.new_fixed_decimal_extractor(start_at, size))
                start_at += size + 1
                continue
            if field.data_type == 'String':
                size = field.size
                self._add_extractor(field.name, DataType.STRING, _extractors.new_string_extractor(start_at, size))
                start_at += size + 1
                continue
            if field.data_type == 'WString':
                size = field.size
                self._add_extractor(field.name, DataType.STRING, _extractors.new_wstring_extractor(start_at, size))
                start_at += (size * 2) + 1
                continue
            if field.data_type == "V_String":
                self._add_extractor(field.name, DataType.STRING, _extractors.new_v_string_extractor(start_at))
                start_at += 4
                self.has_var = True
                continue
            if field.data_type == "V_WString":
                self._add_extractor(field.name, DataType.STRING, _extractors.new_v_wstring_extractor(start_at))
                start_at += 4
                self.has_var = True
                continue
            if field.data_type == "Date":
                self._add_extractor(field.name, DataType.DATE, _extractors.new_date_extractor(start_at))
                start_at += 11
                continue
            if field.data_type == "DateTime":
                self._add_extractor(field.name, DataType.DATE, _extractors.new_date_time_extractor(start_at))
                start_at += 20
                continue
            if field.data_type == "Bool":
                self._add_extractor(field.name, DataType.BOOLEAN, _extractors.new_bool_extractor(start_at))
                start_at += 1
                continue
            if field.data_type == "Byte":
                self._add_extractor(field.name, DataType.BYTE, _extractors.new_byte_extractor(start_at))
                start_at += 2
                continue
            if field.data_type == "Blob" or field.data_type == "SpatialObj":
                self._add_extractor(field.name, DataType.BLOB, _extractors.new_blob_extractor(start_at))
                start_at += 4
                self.has_var = True
                continue
            raise NameError
        self.fixed_size = start_at

    def _add_extractor(self, name: str, data_type: DataType, extractor):
        self._add_field_name_to_index_map(name, data_type)
        self._extractors.append(extractor)

    def _add_field_name_to_index_map(self, name: str, data_type: DataType):
        index = len(self.fields)
        self.fields.append(YxdbField(name, data_type))
        self._name_to_index[name] = index
        return index



from enum import Enum


class DataType(Enum):
    BLOB = 1
    BOOLEAN = 2
    BYTE = 3
    DATE = 4
    DOUBLE = 5
    LONG = 6
    STRING = 7


class YxdbField:
    def __init__(self, name: str, data_type: DataType):
        self.name = name
        self.data_type = data_type



import xml.etree.ElementTree as ET
from io import BytesIO
from typing import List

from yxdb._buffered_record_reader import BufferedRecordReader
from yxdb._metainfo_field import MetaInfoField
from yxdb._yxdb_record import YxdbRecord
from yxdb.yxdb_field import YxdbField

invalid_yxdb_msg = "file is not a valid YXDB format"


class YxdbReader:
    """
    YxdbReader contains the public API for reading YXDB files

    Instantiate YxdbReader with either a string containing the path to a YXDB file,
    or with a BytesIO object containing an in-memory stream of a YXDB file.

    Use the next() method to iterate records.

    Access fields using the following methods:
        * read_index(): read by field index
        * read_name(): read by the name of the field

    The value returned will be a data type appropriate for that field.

    Use the list_fields() method to obtain the list of fields in the YXDB file.
    """

    def __init__(self, *args, **kwargs):
        """
        Instantiate a YXDB reader with 1 of the following parameters:
            * path: a string containing the path to a YXDB file
            * stream: a BytesIO object that streams a YXDB file
        """

        stream: BytesIO = kwargs.get('stream', None)
        path: str = kwargs.get('path', None)
        if stream is None and path is None:
            raise TypeError("either 'path' or 'stream' must be provided")
        if stream is None:
            if not isinstance(path, str):
                raise TypeError("'path' must be a string")
            stream = open(path, 'rb')
        if path is None:
            path = ''

        self._stream = stream
        self._path = path
        self._fields: List[MetaInfoField] = []
        self.num_records = 0
        self._meta_info_size = 0
        self.meta_info_str = ''
        self._record: YxdbRecord = None
        self._record_reader: BufferedRecordReader = None

        try:
            self._load_header_and_meta_info()
        except Exception:
            raise Exception(invalid_yxdb_msg)

    def next(self) -> bool:
        """Returns True if a record is available and False if the end of the file is reached."""

        return self._record_reader.next_record()

    def read_index(self, index: int):
        """Returns the value in a field, specified by the field's index."""

        return self._record.extract_from_index(index, self._record_reader.record_buffer)

    def read_name(self, name: str):
        """Returns the value in a field, specified by the field's name"""

        return self._record.extract_from_name(name, self._record_reader.record_buffer)

    def list_fields(self) -> List[YxdbField]:
        """Provides the list of fields in the YXDB file"""

        return self._record.fields

    def close(self):
        """Closes the YXDB stream early."""

        self._stream.close()

    def _load_header_and_meta_info(self):
        header = self._get_header()
        file_type = str(header[0:21].tobytes(), 'utf_8')
        if "Alteryx Database File" != file_type:
            self._close_stream_and_raise()

        self.num_records = int.from_bytes(header[104:108], 'little')
        self._meta_info_size = int.from_bytes(header[80:84], 'little')
        self._load_meta_info()
        self._record = YxdbRecord(self._fields)
        self._record_reader = BufferedRecordReader(self._stream, self._record.fixed_size, self._record.has_var, self.num_records)

    def _get_header(self) -> memoryview:
        buffer = memoryview(bytearray(512))
        read = self._stream.readinto(buffer)
        if read < 512:
            self._close_stream_and_raise()
        return buffer

    def _load_meta_info(self):
        length = (self._meta_info_size * 2) - 2
        meta_info_bytes = self._stream.read(length)
        self._stream.read(2)
        if len(meta_info_bytes) != length:
            self._close_stream_and_raise()
        self.meta_info_str = str(meta_info_bytes, "utf_16_le")
        self._get_fields()

    def _get_fields(self):
        root = ET.fromstring(self.meta_info_str)
        if root.tag == "RecordInfo":
            record_info = root
        else:
            record_info = root.find("RecordInfo")
        for field in record_info.iter(tag="Field"):
            name = _parse_string(field.get("name"))
            data_type = _parse_string(field.get("type"))
            size = _parse_int(field.get("size"))
            scale = _parse_int(field.get("scale"))

            self._fields.append(MetaInfoField(name, data_type, size, scale))

    def _close_stream_and_raise(self):
        self._stream.close()
        raise IOError(invalid_yxdb_msg)


def _parse_int(value) -> int:
    if value is None:
        return 0
    return int(value)


def _parse_string(value) -> str:
    if value is None:
        raise IOError("YXDB metadata is invalid")
    return value



from io import BytesIO

from yxdb._lzf import Lzf
from yxdb._utility import memview


class BufferedRecordReader:
    def __init__(self, stream: BytesIO, fixed_len: int, has_var_fields: bool, total_records: int):
        self.lzf_buffer_size = 262144
        self.stream = stream
        self.fixed_len = fixed_len
        self.has_var_fields = has_var_fields
        self.total_records = total_records
        if has_var_fields:
            self.record_buffer = memview(fixed_len + 4 + 1000)
        else:
            self.record_buffer = memview(fixed_len)
        self.lzf_in = memview(self.lzf_buffer_size)
        self.lzf_out = memview(self.lzf_buffer_size)
        self.lzf = Lzf(self.lzf_in, self.lzf_out)
        self.lzf_length_buffer = memview(4)
        self.current_record = 0
        self.record_buffer_index = 0
        self.lzf_out_size = 0
        self.lzf_out_index = 0

    def next_record(self) -> bool:
        self.current_record += 1
        if self.current_record > self.total_records:
            self.stream.close()
            return False
        self.record_buffer_index = 0
        if self.has_var_fields:
            self._read_variable_record()
        else:
            self._read(self.fixed_len)
        return True

    def _read_variable_record(self):
        self._read(self.fixed_len + 4)
        var_length = int.from_bytes(self.record_buffer[self.record_buffer_index-4:self.record_buffer_index], "little")
        if self.fixed_len + 4 + var_length > len(self.record_buffer):
            new_length = (self.fixed_len + 4 + var_length) * 2
            new_buffer = memview(new_length)
            new_buffer[:self.fixed_len+4] = self.record_buffer[:self.fixed_len+4]
            self.record_buffer = new_buffer
        self._read(var_length)

    def _read(self, size: int):
        while size > 0:
            if self.lzf_out_size == 0:
                self.lzf_out_size = self._read_next_lzf_block()

            while size + self.lzf_out_index > self.lzf_out_size:
                size -= self._copy_remaining_lzf_out_to_record()
                self.lzf_out_size = self._read_next_lzf_block()
                self.lzf_out_index = 0

            len_to_copy = min(self.lzf_out_size, size)
            self.record_buffer[self.record_buffer_index:self.record_buffer_index+len_to_copy] = self.lzf_out[self.lzf_out_index: self.lzf_out_index+len_to_copy]
            self.lzf_out_index += len_to_copy
            self.record_buffer_index += len_to_copy
            size -= len_to_copy

    def _copy_remaining_lzf_out_to_record(self) -> int:
        remaining_lzf = self.lzf_out_size - self.lzf_out_index
        self.record_buffer[self.record_buffer_index:self.record_buffer_index+remaining_lzf] = self.lzf_out[self.lzf_out_index:self.lzf_out_index+remaining_lzf]
        self.record_buffer_index += remaining_lzf
        return remaining_lzf

    def _read_next_lzf_block(self) -> int:
        lzf_block_length = self._read_lzf_block_length()
        checkbit = lzf_block_length & 0x80000000
        if checkbit > 0:
            lzf_block_length &= 0x7fffffff
            self.stream.readinto(self.lzf_out[:lzf_block_length])
        else:
            read_in = self.stream.readinto(self.lzf_in[:lzf_block_length])
            return self.lzf.decompress(read_in)

    def _read_lzf_block_length(self):
        read = self.stream.readinto(self.lzf_length_buffer)
        if read < 4:
            raise IOError
        return int.from_bytes(self.lzf_length_buffer, "little")



class MetaInfoField:
    def __init__(self, name: str, data_type: str, size: int, scale: int):
        self.name = name
        self.data_type = data_type
        self.size = size
        self.scale = scale



class Lzf:
    def __init__(self, in_bytes: memoryview, out_bytes: memoryview):
        self.in_bytes = in_bytes
        self.out_bytes = out_bytes
        self.iidx = 0
        self.oidx = 0
        self.in_len = 0

    def decompress(self, length: int) -> int:
        self.in_len = length
        self.reset()

        if self.in_len == 0:
            return 0

        while self.iidx < self.in_len:
            ctrl: int = self.in_bytes[self.iidx]
            self.iidx += 1

            if ctrl < 32:
                self._copy_byte_sequence(ctrl)
            else:
                self._expand_repeated_bytes(ctrl)

        return self.oidx

    def reset(self):
        self.iidx = 0
        self.oidx = 0

    def _copy_byte_sequence(self, ctrl: int):
        length: int = ctrl + 1
        if self.oidx + length > len(self.out_bytes):
            raise AttributeError

        self.out_bytes[self.oidx:self.oidx + length] = self.in_bytes[self.iidx: self.iidx + length]
        self.oidx += length
        self.iidx += length

    def _expand_repeated_bytes(self, ctrl: int):
        length: int = ctrl >> 5
        reference: int = self.oidx - ((ctrl & 0x1f) << 8) - 1  # magic

        if length == 7:
            length += self.in_bytes[self.iidx]
            self.iidx += 1

        if self.oidx + length + 2 > len(self.out_bytes):
            raise AttributeError

        reference -= self.in_bytes[self.iidx]
        self.iidx += 1

        length += 2

        while length > 0:
            size = min(self.oidx - reference, length)
            reference = self._copy_from_reference_and_increment(reference, size)
            length -= size

    def _copy_from_reference_and_increment(self, reference: int, size: int) -> int:
        self.out_bytes[self.oidx:self.oidx+size] = self.out_bytes[reference:reference+size]
        self.oidx += size
        return reference + size



import datetime
import struct


def new_bool_extractor(start: int):
    def e(buffer: memoryview):
        value = buffer[start]
        if value == 2:
            return None
        return value == 1
    return e


def new_byte_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+1] == 1:
            return None
        return buffer[start]
    return e


def new_int16_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+2] == 1:
            return None
        return int.from_bytes(buffer[start:start+2], 'little')
    return e


def new_int32_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+4] == 1:
            return None
        return int.from_bytes(buffer[start:start+4], 'little')
    return e


def new_int64_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+8] == 1:
            return None
        return int.from_bytes(buffer[start:start+8], 'little')
    return e


def new_fixed_decimal_extractor(start: int, length: int):
    def e(buffer: memoryview):
        if buffer[start+length] == 1:
            return None
        value = _get_string(buffer, start, length, 1)
        return float(value)
    return e


def new_float_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+4] == 1:
            return None
        return struct.unpack('f', buffer[start:start+4])[0]
    return e


def new_double_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+8] == 1:
            return None
        return struct.unpack('d', buffer[start:start+8])[0]
    return e


def new_date_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+10] == 1:
            return None
        return _parse_date(buffer, start, 10, "%Y-%m-%d")
    return e


def new_date_time_extractor(start: int):
    def e(buffer: memoryview):
        if buffer[start+19] == 1:
            return None
        return _parse_date(buffer, start, 19, "%Y-%m-%d %H:%M:%S")
    return e


def new_string_extractor(start: int, field_len: int):
    def e(buffer: memoryview):
        if buffer[start+field_len] == 1:
            return None
        return _get_string(buffer, start, field_len, 1)
    return e


def new_wstring_extractor(start: int, field_len: int):
    def e(buffer: memoryview):
        if buffer[start+(field_len*2)] == 1:
            return None
        return _get_string(buffer, start, field_len, 2)
    return e


def new_v_string_extractor(start: int):
    def e(buffer: memoryview):
        blob = _parse_blob(buffer, start)
        if blob is None:
            return None
        return str(blob, "utf_8")
    return e


def new_v_wstring_extractor(start: int):
    def e(buffer: memoryview):
        blob = _parse_blob(buffer, start)
        if blob is None:
            return None
        return str(blob, "utf_16_le")
    return e


def new_blob_extractor(start: int):
    def e(buffer: memoryview):
        return _parse_blob(buffer, start)
    return e


def _parse_date(buffer: memoryview, start: int, length: int, fmt: str):
    value = str(buffer[start:start+length].tobytes(), 'utf_8')
    return datetime.datetime.strptime(value, fmt)


def _get_string(buffer: memoryview, start: int, field_len: int, char_size: int):
    end = _get_end_of_string_pos(buffer, start, field_len, char_size)
    if char_size == 1:
        return str(buffer[start:end].tobytes(), 'utf_8')
    else:
        return str(buffer[start:end].tobytes(), 'utf_16_le')


def _get_end_of_string_pos(buffer: memoryview, start: int, field_len: int, char_size: int):
    field_to = start + (field_len * char_size)
    str_len = 0
    i = start
    while i < field_to:
        if buffer[i] == 0 and buffer[i + (char_size-1)] == 0:
            break
        str_len += 1
        i += char_size
    return start + (str_len * char_size)


def _parse_blob(buffer: memoryview, start: int) -> bytes:
    fixed_portion = int.from_bytes(buffer[start:start+4], 'little')
    if fixed_portion == 0:
        return bytearray(0)
    if fixed_portion == 1:
        return None

    if _is_tiny(fixed_portion):
        return _get_tiny_blob(start, buffer)

    block_start = start + (fixed_portion & 0x7fffffff)
    block_first_byte = buffer[block_start]
    if _is_small_block(block_first_byte):
        return _get_small_blob(buffer, block_start)
    else:
        return _get_normal_blob(buffer, block_start)


def _is_tiny(fixed_portion: int) -> bool:
    return fixed_portion & 0x80000000 == 0 and fixed_portion & 0x30000000 != 0


def _get_tiny_blob(start: int, buffer: memoryview) -> bytes:
    value = int.from_bytes(buffer[start:start+4], 'little')
    length = value >> 28
    end = start + length
    return buffer[start:end].tobytes()


def _is_small_block(value) -> bool:
    return (value & 1) == 1


def _get_small_blob(buffer: memoryview, block_start: int) -> bytes:
    first_byte = buffer[block_start]
    blob_len = first_byte >> 1
    blob_start = block_start + 1
    blob_end = blob_start + blob_len
    return buffer[blob_start:blob_end].tobytes()


def _get_normal_blob(buffer: memoryview, block_start: int):
    blob_len = int(int.from_bytes(buffer[block_start:block_start+4], 'little') / 2)
    blob_start = block_start + 4
    blob_end = blob_start + blob_len
    return buffer[blob_start:blob_end].tobytes()
